
from sklearn.feature_extraction import DictVectorizer
from sklearn.svm import LinearSVC
from sklearn.linear_model import Perceptron, LogisticRegression
from sklearn.naive_bayes import MultinomialNB
from sklearn.pipeline import Pipeline

from sklearn.metrics import accuracy_score, confusion_matrix

# Reads the collection of labeled examples, and returns a list of examples and
# a list of corresponding name types ("LOC", "ORG", or "PER").
# Each example consists of a sentence (a list of word/tag pairs) and the
# position of the word we are classifying.
#
# For instance, if we have the sentence "I like Ike!" and our task is to
# classify the name Ike, then the example is represented as follows
#
# ( [ ['I', 'PRP'], ['like', 'VBP'], ['Ike', 'NNP'], ['!', '.'] ], 2 )
#
# and the corresponding name type is 'PER'.
#
def read_examples(filename):
    examples = []
    labels = []
    with open(filename, encoding='utf-8') as f:
        for line in f:
            tokens = line.split()
            examples.append(([ t.split('|') for t in tokens[2:] ], 
                             int(tokens[0])))
            labels.append(tokens[1])
    return examples, labels
            
# extract features from an example (stored as an attribute-value dictionary).
def extract_features(sentence, position):    
    features = {}

    # these are the features that we added during the lecture    
    if position == 0:
        features['previous_word'] = 'FIRST_WORD'
    else:
        features['previous_word'] = sentence[position-1][0]
    if position == 0:
        features['previous_pos'] = 'FIRST_WORD'
    else:
        features['previous_pos'] = sentence[position-1][1]

    # ... and here are some additional ones.
    features['word'] = sentence[position][0]
    features['pos'] = sentence[position][1]
    if position < len(sentence)-1:
        features['word_after'] = sentence[position+1][0]
        features['pos_after'] = sentence[position+1][1]
    else:
        features['word_after'] = 'END'
        features['pos_after'] = 'END'

    return features


# read the examples and corresponding NE labels
examples, Y = read_examples('ne_examples.txt')

# extract features from each example
X = [ extract_features(s[0], s[1]) for s in examples ]

# split into training and test sets
split_point = int(0.75*len(X))
Xtrain = X[:split_point]
Ytrain = Y[:split_point]
Xtest = X[split_point:]
Ytest = Y[split_point:]

# train a pipeline consisting of a DictVectorizer (since our features are
# stored as attribute-value pairs) followed by a Perceptron classifier
print('Training NE classifier...')
classifier = Pipeline([('vec', DictVectorizer()),
                       ('cls', Perceptron())])
classifier.fit(Xtrain, Ytrain)

# evaluate
print('Evaluating on the test set...')
Yguess = classifier.predict(Xtest)

# print the accuracy
print('Classification accuracy: {:.4f}'.format(accuracy_score(Ytest, Yguess)))

# print a confusion matrix
class_list = sorted(set(Ytrain + Ytest))
print('Confusion matrix (classes = {0})'.format(" ".join(class_list)))
print(confusion_matrix(Yguess, Ytest, class_list))
